/*=============================================================================
vitalnix/C-spark - User Management UI
main.c - Startup code
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 1999 - 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <sys/file.h> /* lockf() */
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <popt.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libHX.h>
#include <accdb.h>
#include "cui/global.h"
#include "cui/xml_in.h"

static int main2(int, char **);
static void init_var(int, char **);
static void deinit_var(void);
static void init_applock(void);

static int count_args(const char **);
static int parse_options(int *, const char ***);
static int load_config(char *);
static void exit_pg(signed int, char *, ...);

struct optmap_s Optmap;

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    return !main2(argc, argv);
}

static int main2(int argc, char **argv) {
    int eax;
    init_var(argc, argv);
    if(!(eax = parse_options(&argc, (const char ***)&argv))) { return !eax; }
    init_applock();

    while(1) {
        char buf[MAXLNLEN];
        int c;

        printf(
          "Main menu\n"
          "---------\n"
          "<1>  Import from XML\n"
          "<9>  Quit\n"
          "\n"
          "CHOICE>"
        );
        fgets(buf, MAXLNLEN, stdin);
        c = strtoul(buf, NULL, 0);
        if(c == 1) {
            if(db_open()) { apply_ds(); }
            db_close();
        }
        if(c == 9) {
            break;
        }
    }

    deinit_var();
    exit_pg(!1, NULL);
    return !1;
}

static void init_var(int argc, char **argv) {
    char *p;
    memset(&Optmap, '\0', sizeof(Optmap));
    Optmap.progv = "v1.90.8.72";
    Optmap.progd = __DATE__ " " __TIME__;

    if(!isatty(fileno(stdin)) && !isatty(fileno(stdout)) &&
     (p = getenv("DISPLAY")) != NULL && p[0] != '\0') {
        int r;
        r = HX_fsystem(HX_FSYSTEM_ARGV | HX_FSYSTEM_EXEC, "xterm", argv,
         "-geometry", "80x25", "-bg", "#002040", "-fg", "gray", "-e", NULL);
        if(r != 0) {
            fprintf(stderr, "HX_fsystem() unsuccessful: %s\n",
             strerror(errno));
            exit(1);
        }
    }

    HX_strclone(&Optmap.f_lock, "/var/run/c-spark.pid");
    HX_strclone(&Optmap.f_skeld, "");
    HX_strclone(&Optmap.p_pager, "/usr/bin/less");
    HX_strclone(&Optmap.pa_pager, "-MSi %1$s");
    load_config("/etc/spark.cfg");
    load_config("vetc/spark.cfg");
    load_config("spark.cfg");

    printf(
      "=====[\e[1m C-spark %s <devel> (%s) \e[0m]=====\n\n" // ]]
      "This program comes with ABSOLUTELY NO WARRANTY; it is free software,\n"
      "you are welcome to redistribute it under certain conditions;\n"
      "for details see the files doc/FAL.txt and doc/GPL2.txt which should\n"
      "have come with this program.\n\n",
      Optmap.progv, Optmap.progd
    );

    umask(0077);
    return;
}

static void deinit_var(void) {
    HX_strclone(&Optmap.f_lock, NULL);
    HX_strclone(&Optmap.f_skeld, NULL);
    HX_strclone(&Optmap.p_pager, NULL);
    HX_strclone(&Optmap.pa_pager, NULL);
    unlink(Optmap.f_lock);
    return;
}

static void init_applock(void) {
    char buf[16];
    int fd;

    if((fd = open(Optmap.f_lock, O_RDWR | O_CREAT, 0666)) < 0) {
        exit_pg(1 | P_ERR, "open()");
    }

    if(lockf(fd, F_TLOCK, 0) < 0) {
        exit_pg(1 | P_ERR, "C-spark already running?\nlockf()");
    }

    ftruncate(fd, 0);
    snprintf(buf, 8, "%d\n", getpid());
    write(fd, buf, strlen(buf));
    fcntl(fd, F_SETFD, FD_CLOEXEC);
    // never close that file or the lock will be gone
    return;
}

static int count_args(const char **args) {
    int argk = 0;
    while(*args++ != NULL) { ++argk; }
    return argk;
}

static int parse_options(int *argc, const char ***argv) {
    static const char *_empty_argv[] = {NULL};
    char *cfg;
    struct poptOption options_table[] = {
        {"config", 'c', POPT_ARG_STRING, &cfg, 'c',
         "Load alternate config file", NULL},
        {NULL, 'h', POPT_ARG_NONE, NULL, 'h', "Show this help message", NULL},
        {"version", 'v', POPT_ARG_NONE, NULL, 'v',
         "Show version number and exit", NULL},
        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
    while((c = poptGetNextOpt(ctx)) >= 0) {
        switch(c) {
            case 'c':
                if(!load_config(cfg)) {
                     fprintf(stderr, "Parsing of %s unsuccessful\n", cfg);
                }
                break;
            case 'h':
                poptPrintHelp(ctx, stdout, 0);
                exit(0);
                break;
            case 'v':
                printf("C-spark %s (%s)\n", Optmap.progv, Optmap.progd);
                exit(0);
                break;
        }
    }

    if(c < -1) {
        fprintf(stderr, "%s: %s\n", poptBadOption(ctx, 0), poptStrerror(c));
        poptFreeContext(ctx);
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
        *argc = 0;
        *argv = _empty_argv;
    }

    return 1;
}

static int load_config(char *fn) {
    static struct rconfig_opt options_table[] = {
        {"PAGER", RCONF_STRING, &Optmap.p_pager, NULL},
        {"PAGER_ARGS", RCONF_STRING, &Optmap.pa_pager, NULL},
        {"SKEL", RCONF_STRING, &Optmap.f_skeld, NULL},
        {"SPLIT_HOME", RCONF_BOOL, &Optmap.b_split, NULL},
        {NULL},
    };

    return accdb_rconfig(fn, options_table);
}

static void exit_pg(signed int v, char *fmt, ...) {
    if(fmt != NULL) {
        if(v & P_ERR) { perror(fmt); }
        else {
            va_list argp;
            va_start(argp, fmt);
            vfprintf(stderr, fmt, argp);
            va_end(argp);
        }
    }
    exit(v & ~P_ERR);
    return;
}

//==[ End of file ]============================================================
