/*=============================================================================
vitalnix/C-spark - User Management UI
user.c - Where the real work happens
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 1999 - 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "cui/ds_entry.h"
#include "cui/global.h"
#include "cui/xml_in.h"

#define mm_sysctl(req, args...) \
     if(mm_output->sysctl != NULL) { \
         mm_output->sysctl(req, mm_output->state, ## args); \
     }

static int read_input(void);
static int get_dgrp(struct accdb_group *);

static void cleanup_storage_addlist(void);
inline static void cleanup_storage_dellist(void);
static void free_addlist_entry(struct ds_entry *);
inline static void handle_user(struct accdb_user *, size_t *);

static struct HXbtree *storage_addlist = NULL;
static struct HXlld *storage_dellist = NULL;
static struct accdb_module *mm_output = NULL;

//-----------------------------------------------------------------------------
int db_open(void) {
    int eax;

    if((mm_output = accdb_load("shadow")) == NULL) {
        fprintf(stderr, "Could not load ACCDB back-end: %s\n",
         strerror(errno));
        return 0;
    }

    if((eax = mm_output->open(&mm_output->state)) <= 0) {
        fprintf(stderr, "Could not open ACCDB back-end: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        accdb_unload(mm_output);
        return 0;
    }

    return 1;
}

void db_close(void) {
    if(mm_output != NULL) {
        if(mm_output->state != NULL) { mm_output->close(mm_output->state); }
        accdb_unload(mm_output);
    }
    return;
}

void apply_ds(void) {
    struct HXbtree *lnlist;
    struct accdb_group dgroup;

    cleanup_storage_addlist();
    cleanup_storage_dellist();

    if(!get_dgrp(&dgroup)) { return; }
    if(read_input() <= 0) { return; }

    { /* And this is now the comparison algorithm, (half of) the core. While
      traversing the users, we mark used login names by putting them in an
      array (lnlist). Login names of users which are about to be deleted are
      also pushed there, so they do not get reassigned right now, but at the
      next run-through. */
      size_t users_grp = 0, users_keep = 0;
      struct accdb_user passwd;

      lnlist = HXbtree_init(HXBT_CPTR | HXBT_PCMP, strcmp);
      storage_dellist = HXlld_init(0);

      while(mm_output->usertrav(mm_output->state, &passwd)) {
          HXbtree_add(lnlist, passwd.lname);

          if((dgroup.gid == -1 || dgroup.gid == passwd.gid) &&
           (dgroup.gname == NULL || passwd.lname == NULL ||
           strcmp(dgroup.gname, passwd.lname) == 0)) {
              // Users we look for (only specified group)
              ++users_grp;
              handle_user(&passwd, &users_keep);
          }
      }

      printf("%u group member%s found\n", users_grp,
       users_grp != 1 ? "s" : "");
      printf("\t %u to keep\n", users_keep);
      printf("\t %lu to delete\n", storage_dellist->itemcount);
      printf("%lu to add\n", storage_addlist->itemcount);
    }

    { // Here comes the fixup loop (other part of the core).
      struct HXbtrav *travp = HXbtrav_init(storage_addlist, NULL);
      struct HXbtree_node *node;

      while((node = HXbtraverse(travp)) != NULL) {
          struct ds_entry *entry = node->data;
          char buf[16];
          int i = 0;
          buf[15] = '\0';
          strncpy(buf, entry->lname, 15);

          /* If the fixup'ed login name with the unique number already exists,
          bump the unique number. */
          while(HXbtree_find(lnlist, buf) != NULL) {
              fixup_login_name(entry->lname, ++i, buf, 16);
          }

          if(strcmp(entry->lname, buf) != 0) {
              HX_strclone(&entry->lname, buf);
          }
          HXbtree_add(lnlist, buf);
      }

      HXbtrav_free(travp);
    }

    HXbtree_free(lnlist);
    mm_sysctl(ACCDB_ADDFLAGS, ACCDB_NOAUTOFLUSH);

    // Adding the users to the system database
    if(storage_addlist->itemcount > 0) {
        struct HXbtrav *travp = HXbtrav_init(storage_addlist, NULL);
        struct HXbtree_node *node;

        while((node = HXbtraverse(travp)) != NULL) {
            struct ds_entry *in = node->data;
            struct accdb_user out;

            memset(&out, '\0', sizeof(struct accdb_user));
            out.lname = in->lname;
            out.uid   = -1;
            out.gid   = dgroup.gid;
            out.gecos = node->key;
            out.home  = NULL;
            out.shell = NULL;
            out.igrp  = dgroup.gname;
            out.sgrp  = NULL;

            out.passwd      = "!";
            out.last_change = 0;
            out.keep_min    = 0;
            out.keep_max    = 366;
            out.warn_age    = 21;
            out.expire      = -1;
            out.inactive    = -1;
            mm_output->useradd(mm_output->state, &out);
            printf("Added %s as %s\n", out.gecos, out.lname);
        }

        HXbtrav_free(travp);
    }

    // Deleting the old users
    if(storage_dellist->itemcount > 0) {
        struct HXlld_node *travp = storage_dellist->first;

        while(travp != NULL) {
            printf("delete %s\n", (char *)travp->ptr);
            travp = travp->Next;
        }
    }

    mm_sysctl(ACCDB_FLUSHDB);
    cleanup_storage_addlist();
    cleanup_storage_dellist();
    return;
}

static int get_dgrp(struct accdb_group *grp) {
    /* Here, we request a group to which compare the data source against.
    Usually, a group is always the same for a particular input set otherwise
    strange things may happen. You can either enter the group name or a GID.
    (Once there is a non-digit in your input, it is treated as a group
    name.) */
    char buf[MAXLNLEN];

    printf("Destianation group> ");
    fgets(buf, MAXLNLEN, stdin);
    HX_chomp(buf);
    if(strlen(buf) < 1) { return 0; }

    if(only_digits(buf)) {
        grp->gid = strtol(buf, NULL, 0);
        grp->gname = NULL;
    } else {
        grp->gid = -1;
        grp->gname = buf;
    }

    if(!mm_output->groupinfo(mm_output->state, grp, grp, 1)) {
        printf("Group \"%s\" not found\n", buf);
        return 0;
    }

    return 1;
}

static int read_input(void) {
    /* In this part, we stuff all users from the data source into an
    associative array, called storage_addlist. Some elements of that array will
    further be changed until it is ready for running through apply_ds(). */
    struct ds_entry *bucket;
    char buf[MAXLNLEN];
    void *state;
    int eax;

    printf("FIND CHANGES; Input file> ");
    fgets(buf, MAXLNLEN, stdin);
    HX_chomp(buf);

    if((eax = xmlds_open(buf, &state)) <= 0) { return eax; }

    /* This loop fills the storage_addlist assoc array by generating a suited
    (pre-stage) login name.
    Pre-stage login names: these login names may already be used by the system
    and must first be fixed up (i.e. by adding a unique postcount number), so
    that there is no double login name in the user database. */

    storage_addlist = HXbtree_init(HXBT_ASSOC);

    while(1) {
        bucket = malloc(sizeof(struct ds_entry));
        if(xmlds_read(state, bucket) <= 0) {
            free(bucket);
            break;
        }

        bucket->lname = make_login_name(bucket->nname, bucket->vname);

        if(bucket->nname != NULL) {
            snprintf(buf, MAXLNLEN, "%s %s", bucket->vname, bucket->nname);
            bucket->rname = HX_strdup(buf);
        } else {
            bucket->rname = HX_strdup(bucket->vname);
        }

        if(bucket->xuid != NULL) {
            snprintf(buf, MAXLNLEN, "%s,%s", bucket->rname, bucket->xuid);
        } else {
            buf[MAXLNLEN - 1] = '\0';
            strncpy(buf, bucket->rname, MAXLNLEN - 1);
        }
        HXbtree_add(storage_addlist, buf, bucket);
    }

    xmlds_close(state);
    return 1;
}

//-----------------------------------------------------------------------------
static void cleanup_storage_addlist(void) {
    if(storage_addlist != NULL) {
        struct HXbtrav *travp = HXbtrav_init(storage_addlist, NULL);
        struct HXbtree_node *node;

        while((node = HXbtraverse(travp)) != NULL) {
            free_addlist_entry(node->data);
        }

        HXbtrav_free(travp);
        HXbtree_free(storage_addlist);
        storage_addlist = NULL;
    }
    return;
}

inline static void cleanup_storage_dellist(void) {
    if(storage_dellist != NULL) {
        HXlld_free(storage_dellist);
        storage_dellist = NULL;
    }
    return;
}

static void free_addlist_entry(struct ds_entry *e) {
    if(e == NULL) { return; }
    free(e->vname);
    if(e->nname != NULL) { free(e->nname); }
    if(e->xuid != NULL) { free(e->xuid); }
    if(e->sgroup != NULL) { free(e->sgroup); }
    free(e->lname);
    free(e->rname);
    free(e);
    return;
}

inline static void handle_user(struct accdb_user *p, size_t *ukeep) {
    char buf[MAXLNLEN], *brkp;

    buf[511] = '\0';
    strncpy(buf, p->gecos, 511);

    if((brkp = strchr(buf, ',')) != NULL &&
     (brkp = strchr(buf, ',')) != NULL) {
        *brkp = '\0';
    }

    if(HXbtree_find(storage_addlist, buf) != NULL) {
        ++*ukeep;
        free_addlist_entry(HXbtree_get(storage_addlist, buf));
        HXbtree_del(storage_addlist, buf);
        // what remains in storage_addlist is the users to add
    } else {
        // not anymore in data source
        HXlld_push(storage_dellist, p->lname);
    }

    return;
}

//==[ End of file ]============================================================
